{% for pin in pinned_history %}
- name: "pin_{{ loop.index }}"
  content: |
    {{ settings.start_token }}{{ escape_special_characters(pin) }}{{ settings.eom }}
  {% set settings.start_token = settings.bom %}
{% endfor %}
